A mail system based on JMS and MQSeries

Chris Markes
HCI Architect
IBM UK Laboratories (UK)   c C Markes 2000



#!/bin/ksh

JMS_HOME=/usr/mqm/java
JMS_LIB=$JMS_HOME/lib
JMS_CLASSPATH=/usr/jdk_base/lib/classes.zip:\
$JMS_LIB:\
$JMS_LIB/com.ibm.mqjms.jar:\
$JMS_LIB/jms.jar:\
$JMS_LIB/jndi.jar:\
$JMS_LIB/fscontext.jar:\
$JMS_LIB/ldap.jar:\
$JMS_LIB/providerutil.jar:\
$JMS_LIB/com.ibm.mq.jar:\
$JMS_LIB/com.ibm.mqbind.jar:.
FACTORY=com.sun.jndi.fscontext.RefFSContextFactory
REPOSITORY=file:$JMS_HOME/directory

echo Starting JMSMail...

java -classpath $JMS_CLASSPATH \
-Djava.naming.factory.initial=$FACTORY \
-Djava.naming.provider.url=$REPOSITORY \
JMSMail $@


@echo off

SET JMS_HOME=\MQSeries\java
SET JMS_LIB=%JMS_HOME%\lib
SET JMS_CLASSPATH=%JMS_LIB%;%JMS_LIB%\com.ibm.mqjms.jar;%JMS_LIB%\jms.jar;%JM S_LIB%\jndi.jar;%JMS_LIB%\fscontext.jar;%JMS_LIB%\ldap.jar;%JMS_LIB%\prov iderutil.jar;%JMS_LIB%\com.ibm.mq.jar;%JMS_LIB%\com.ibm.mqbind.jar;.
SET FACTORY=com.sun.jndi.fscontext.RefFSContextFactory
SET REPOSITORY=file:/MQSeries/mqm/java/directory

echo Starting JMSMail...

java -classpath %JMS_CLASSPATH% -Djava.naming.factory.initial=%FACTORY% Djava.naming.provider.url=%REPOSITORY% JMSMail %1 %2 %3 %4 %5




import java.io.*;
import java.util.*;
import javax.naming.*;
import javax.jms.*;

/**
* JMS sample application modelling a simple mail system. The
* user interface is loosely based on Unix's common mail client.
*/

public class JMSMail implements MessageListener, ExceptionListener
{
   final static String NEWLINE =
      File.separatorChar == '\\' ? "\r\n" : "\n";

   static InitialContext context;

   String username;
   QueueConnection connection;
   QueueSession session;
   Queue mailbox;
   Hashtable table = new Hashtable ();
   BufferedReader input = new BufferedReader
      (new InputStreamReader (System.in));

   /**
    * Main application's entry point. If any command-line arguments
    * are supplied, the application runs in send mode, so that a
    * message is composed and sent to each of the specified
    * addressees in turn. With no command-line arguments, the
    * application runs in receive mode and manipulates received
    * messages in the user's mailbox using simple commands.
    */

   public static void main (String [] args)
   {
      try
      {
         // Set up the JNDI context.

         context = new InitialContext (System.getProperties ());

         // The username identifies the 'from' address to use when
         // sending and the mailbox (queue) to use when receiving.

         String username = System.getProperty ("jmsmail.user.name",
            System.getProperty ("user.name"));
         JMSMail app = new JMSMail (username);

         if (args.length > 0)
         {
            // Compose a message and send it to each addressee
            // specified on the command line.

            Message message = app.composeMessage (null);

            for (int i = 0; i < args.length; i++)
            {
               try
               {
                  app.sendMessage (message, args [i]);
               }

               catch (NamingException ne)
               {
                  System.err.println (args [i] +
                     ": JNDI lookup error: " + ne);
               }
            }
         }
         else
         {
            // Run in receive mode operating on messages in the
            // user's mailbox.

            app.run ();
         }

         app.disconnect ();
      }

      catch (JMSException jmse)
      {
         System.err.println ("JMS error: " + jmse +
            " (" + jmse.getLinkedException () + ')');
      }

      catch (NamingException ne)
      {
         System.err.println ("Invalid destination: " + ne);
      }

      catch (IOException ioe)
      {
         System.err.println ("Error reading input: " + ioe);
      }
   }

   /**
    * Returns a connection using a named ConnectionFactory.
    */

   static QueueConnection getConnection (String name) throws
      JMSException, NamingException
   {
      QueueConnectionFactory factory =
         (QueueConnectionFactory) context.lookup (name);

      return factory.createQueueConnection ();
   }

   /**
    * Returns a reference to a named queue.
    */

   static Queue getDestination (String name) throws
      JMSException, NamingException
   {
      return (Queue) context.lookup (name);
   }

   /**
    * Constructor for a JMSMail client (sender or receiver).
    */

   public JMSMail (String username) throws
      JMSException, NamingException
   {
      this.username = username;

      // We normally use ConnectionFactory 'jmsmail', though this
      // can be overridden using the jmsmail.server.name property.

      connection = getConnection (System.getProperty
         ("jmsmail.server.name", "jmsmail"));
      connection.setExceptionListener (this);
      session = connection.createQueueSession
         (false, Session.AUTO_ACKNOWLEDGE);
      mailbox = getDestination (username);
   }

   /**
    * Compose a JMS message from the command line. Text entry is
    * terminated either by an EOF (CTRL-D) or a line containing
    * only a period ('.').
    */

   public Message composeMessage (String subject) throws
      IOException, JMSException
   {
      TextMessage message = session.createTextMessage ();
      StringBuffer buffer = new StringBuffer ();
      String line;

      try
      {
         System.out.print ("Subject: ");

         if (subject == null)
         {
            subject = input.readLine ();
         }
         else
         {
            System.out.println (subject);
         }

         while ((line = input.readLine ()) != null &&
            !line.equals ("."))
         {
            buffer.append (line);
            buffer.append (NEWLINE);
         }
      }

      catch (EOFException eofe)
      {
         ;
      }

      message.setText (new String (buffer));

      // Set message properties; 'from' and 'subject' are
      // application-defined.

      message.setJMSReplyTo (mailbox);
      message.setStringProperty ("from", username);
      message.setStringProperty ("subject", subject);

      return message;
   }

   /**
    * Send a JMS message to a named destination.
    */

   public void sendMessage (Message message, String destination)
      throws JMSException, NamingException
   {
      Queue queue;
      QueueSender sender;

      try
      {
         queue = getDestination (destination);
      }

      catch (NullPointerException npe)
      {
         throw new NamingException ("Invalid destination: " +
            destination);
      }

      sender = session.createSender (queue);
      sender.send (message);
      sender.close ();
   }

   /**
    * Finds the JMSMessageID property of the message in the
    * mailbox with the specified index (in receive mode).
    */

   String getMessageID (int index) throws
      ArrayIndexOutOfBoundsException
   {
      String messageID = (String) table.get (new Integer (index));

      if (messageID == null)
      {
         throw new ArrayIndexOutOfBoundsException
            ("Invalid message number " + index + '.');
      }
      else
      {
         return messageID;
      }
   }

   /**
    * Return a string listing the sender, timestamp, and subject
    * along with the index of the messages in the mailbox
    * (receive mode).
    */

   public String listMessages () throws JMSException
   {
      QueueBrowser browser = session.createBrowser (mailbox);
      Enumeration e = browser.getEnumeration ();
      StringBuffer sb = new StringBuffer ();
      int i = 0;

      synchronized (table)
      {
         table.clear ();

         while (e.hasMoreElements ())
         {
            Message message = (Message) e.nextElement ();
            String subject = message.getStringProperty ("subject");

            sb.append (++i);
            sb.append ("  ");
            sb.append (message.getStringProperty ("from"));
            sb.append ("  ");
            sb.append (new Date (message.getJMSTimestamp ()));
            sb.append ("  ");
            sb.append (subject == null ? "" : subject);
            sb.append (NEWLINE);

            table.put (new Integer (i), message.getJMSMessageID ());
         }
      }

      sb.append (i);
      sb.append (i == 1 ? " message." : " messages.");
      browser.close ();

      return new String (sb);
   }

   /**
    * Returns the message with the specified JMSMessageID from the
    * mailbox (receive mode).
    */

   public Message browseMessage (String messageID) throws
      JMSException
   {
      QueueBrowser browser = session.createBrowser (mailbox);
      Enumeration e = browser.getEnumeration ();
      Message message = null;

      while (e.hasMoreElements () && message == null)
      {
         message = (Message) e.nextElement ();

         if (!message.getJMSMessageID ().equals (messageID))
         {
            message = null;
         }
      }

      browser.close ();

      return message;
   }

   /**
    * Returns the message with the specified JMSMessageId from
    * the mailbox, deleting it from the queue.
    */

   public Message receiveMessage (String messageID) throws
      JMSException
   {
      String selector = "JMSMessageID='" + messageID + '\'';
      QueueReceiver receiver = session.createReceiver
         (mailbox, selector);
      Message message = receiver.receiveNoWait ();

      receiver.close ();

      return message;
   }

   /**
    * This MessageListener method is called when a new message is
    * received. The message is passed directly, and is not retained
    * in the mailbox queue.
    */

   public void onMessage (Message message)
   {
      System.out.print ((char) 0x07);   // beep

      try
      {
         System.out.println ("\nNew mail from " +
            message.getStringProperty ("from").trim () + ": ");
      }

      catch (JMSException jmse)
      {
         System.out.println ("\nNew mail: ");
      }

      System.out.println (message);
      System.out.print ("? ");
   }

   /**
    * This ExceptionListener method is called if a problem occurs
    * with the connection to the JMS service provider while we're
    * not looking.
    */

   public void onException (JMSException jmse)
   {
      System.err.println ("JMS connection error: " + jmse +
         " (" + jmse.getLinkedException () + ')');
      System.exit (-1);
   }

   /**
    * Runs this JMSMail instance as a receiver, reading mailbox
    * commands from the command line. For a list of commands,
    * see the help method.
    */

   public void run () throws JMSException, IOException
   {
      String selector = "JMSTimestamp > " +
         (int) System.currentTimeMillis ();
      QueueSession listenerSession = connection.createQueueSession
         (false, Session.AUTO_ACKNOWLEDGE);
      QueueReceiver listenerReceiver =
         listenerSession.createReceiver (mailbox, selector);
      String command;
      int index = 0;
      boolean quit = false;

      listenerReceiver.setMessageListener (this);
      System.out.println (listMessages ());
      connection.start ();

      do
      {
         System.out.print ("? ");

         try
         {
            if ((command = input.readLine ().trim ()).equals (""))
            {
               System.out.println
                  (browseMessage (getMessageID (++index)));
            }
            else if (command.equals ("l"))
            {
               System.out.println (listMessages ());
               index = 0;
            }
            else if (command.equals ("d"))
            {
               receiveMessage (getMessageID (index));
               System.out.println
                  ("Message " + index++ + " deleted.");
            }
            else if (command.equals ("r"))
            {
               Message message = browseMessage (getMessageID (index));

               sendMessage (composeMessage ("Re: " +
                  message.getStringProperty ("subject")),
                     message.getStringProperty ("from"));
               System.out.println ("Reply sent.");
            }
            else if (command.equals ("s"))
            {
               System.out.print ("To: ");

               String to = input.readLine ();

               sendMessage (composeMessage (null), to);
               System.out.println ("Message sent.");
            }
            else if (command.equals ("q"))
            {
               quit = true;
            }
            else
            {
               try
               {
                  index = Integer.parseInt (command);
                  System.out.println
                     (browseMessage (getMessageID (index)));
               }

               catch (NumberFormatException nfe)
               {
                  System.out.println (help ());
               }
            }
         }

         catch (ArrayIndexOutOfBoundsException abe)
         {
            System.out.println (abe.getMessage ());
         }

         catch (NamingException ne)
         {
            System.out.println ("Invalid destination");
         }
      }
      while (!quit);

      listenerReceiver.close ();
      listenerSession.close ();
   }

   /**
    * Displays a list of valid commands.
    */

   public String help ()
   {
      return "Commands: " +
         "\n\t<n>: show message number n (sets index to n)" +
         "\n\tl: list messages (resets index)" +
         "\n\td: delete message at current index" +
         "\n\tr: reply to message at current index" +
         "\n\ts: send new message" +
         "\n\tq: quit";
   }

   /**
    * Terminates the connection to the JMS service provider.
    */

   public void disconnect () throws JMSException
   {
      session.close ();
      connection.close ();
   }
}


